// Turn off the computer after some time. Won't work under Win95/98 since
// ExitWindowsEx can't be called from a console application.

#include <windows.h>
#include <tlhelp32.h>
#include <stdio.h>

// This was taken from a Microsoft(c) sample. It turns off the explorer process
// before the shutdown.
//------------------------------------------------------------------------------
BOOL WINAPI ExitWindows9x(DWORD dwFlags)
//------------------------------------------------------------------------------
{
   if (dwFlags & EWX_FORCE)
   {
   // Enumerate and terminate all processes named Explorer. This code calls 
   // the TerminateProcess API, which is usually not recommended. We must  
   // use this here, however, because Explorer's auto-restart will cause       
   // logoff problems.

      HINSTANCE      hInstLib;
      HANDLE         hSnapShot;
      PROCESSENTRY32 procentry;
      BOOL           bFlag;

      // ToolHelp function pointers.
      HANDLE (WINAPI *lpfCreateToolhelp32Snapshot)(DWORD,DWORD);
      BOOL (WINAPI *lpfProcess32First)(HANDLE,LPPROCESSENTRY32);
      BOOL (WINAPI *lpfProcess32Next)(HANDLE,LPPROCESSENTRY32);


      hInstLib = LoadLibraryA( "Kernel32.DLL" );
      if( hInstLib == NULL )
       return FALSE;

      // Get procedure addresses.
      // We are linking to these functions of Kernel32
      // explicitly, because otherwise a module using
      // this code would fail to load under Windows NT,
      // which does not have the Toolhelp32
      // functions in the Kernel 32.
      lpfCreateToolhelp32Snapshot=
                     (HANDLE(WINAPI *)(DWORD,DWORD))
                     GetProcAddress( hInstLib,
                     "CreateToolhelp32Snapshot" );
      lpfProcess32First=
                     (BOOL(WINAPI *)(HANDLE,LPPROCESSENTRY32))
                     GetProcAddress( hInstLib, "Process32First" );
      lpfProcess32Next=
                     (BOOL(WINAPI *)(HANDLE,LPPROCESSENTRY32))
                     GetProcAddress( hInstLib, "Process32Next" );

      if( lpfProcess32Next == NULL ||
          lpfProcess32First == NULL ||
          lpfCreateToolhelp32Snapshot == NULL )
      {
         FreeLibrary( hInstLib );
         return FALSE;
      }

      // Get a handle to a ToolHelp snapshot of the systems
      // processes.
      hSnapShot = lpfCreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0 );
      if( hSnapShot == INVALID_HANDLE_VALUE )
      {
         FreeLibrary( hInstLib );
         return FALSE;
      }

      // Get the first process's information.
      procentry.dwSize = sizeof(PROCESSENTRY32);
      bFlag = lpfProcess32First( hSnapShot, &procentry );

      // While there are processes, keep looping.
      while( bFlag )
      {
         {  // We need to see if the app is named Explorer.exe.
            HANDLE hProcess;
            INT nPos;

            nPos = lstrlen(procentry.szExeFile);

            if (nPos)
            {
               while (procentry.szExeFile[--nPos] != '\\' );
               if(!lstrcmpi("explorer.exe",&(procentry.szExeFile[nPos+1])))
               {
                  // Terminate the process.
                  hProcess = OpenProcess(PROCESS_ALL_ACCESS,FALSE,
                                                 procentry.th32ProcessID);
                  TerminateProcess(hProcess, 1);
                  CloseHandle(hProcess);
               }

            }
          }

         procentry.dwSize = sizeof(PROCESSENTRY32) ;
         bFlag = lpfProcess32Next( hSnapShot, &procentry );
      }

      CloseHandle(hSnapShot);

      // Free the library.
      FreeLibrary( hInstLib ) ;
   }

   return ExitWindowsEx(dwFlags,0);
}

// Returns true for Windows NT/2000/XP
//------------------------------------------------------------------------------
bool isWindowsNt()
//------------------------------------------------------------------------------
{
    return GetVersion() < 0x80000000;
}

// Process needs SE_SHUTDOWN_NAME privilege to turn off the computer.
//------------------------------------------------------------------------------
void enableShutDownPrivilege()
//------------------------------------------------------------------------------
{
    HANDLE hToken; 
    TOKEN_PRIVILEGES tkp; 
 
    // Get a token for this process. 
 
    if (!OpenProcessToken(GetCurrentProcess(), 
                          TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken)) 
    {
        MessageBox(0, "OpenProcessToken", "Error", MB_OK); 
    }
 
    // Get the LUID for the shutdown privilege. 
 
    LookupPrivilegeValue(NULL, SE_SHUTDOWN_NAME, 
                         &tkp.Privileges[0].Luid); 
 
    tkp.PrivilegeCount = 1;  // one privilege to set    
    tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED; 
 
    // Get the shutdown privilege for this process. 
 
    AdjustTokenPrivileges(hToken, FALSE, &tkp, 0, 
                          (PTOKEN_PRIVILEGES)NULL, 0); 
 
    // Cannot test the return value of AdjustTokenPrivileges. 
 
    if (GetLastError() != ERROR_SUCCESS) 
    {
        MessageBox(0, "AdjustTokenPrivileges", "Error", MB_OK); 
    }
}


//------------------------------------------------------------------------------
int turnOff(bool force)
//------------------------------------------------------------------------------
{
    if (isWindowsNt())
    {
        enableShutDownPrivilege();
    }

    if (force)
    {
        if (isWindowsNt())
        {
            ExitWindowsEx(EWX_POWEROFF|EWX_FORCE, 0);
        }
        else
        {
            ExitWindows9x(EWX_POWEROFF|EWX_FORCE);
        }
    }
    else
    {
        ExitWindowsEx(EWX_POWEROFF, 0);
    }

    return 0;
}

//------------------------------------------------------------------------------
void displayUsage()
//------------------------------------------------------------------------------
{
    printf("\nUsage: turnoff [ -f | -t xx]\n\n"
           "\t-f\tforce applications to close without warning\n"
           "\t-t xx\twait xx seconds until turning off the computer\n\n");
}

//------------------------------------------------------------------------------
char getArgument(const char* string)
//------------------------------------------------------------------------------
{
    if (string[0] == '-')
    {
        return string[1];
    }
    else
    {
        return 0;
    }
}

//------------------------------------------------------------------------------
int main(int argc, char* argv[])
//------------------------------------------------------------------------------
{
    bool ok = true;
    if (argc > 4)
    {
        ok = false;
    }

    bool force = false;
    if (ok && argc > 1)
    {
        for (int i = 1; i<argc; ++i)
        {
            switch (getArgument(argv[i]))
            {
            case 'f':
                force = true;
                break;

            case 't':
                {
                    int timeout;
                    if (i<argc-1 && argv[i+1][0] != '-')
                    {
                        timeout = atoi(argv[i+1]);
                        Sleep(timeout*1000);
                        ++i;
                    }
                    else
                    {
                        ok = false;
                    }
                }
                break;

            default:
                ok = false;
                break;
            }
        }
    }
 
    if (ok)
    {
        return turnOff(force);
    }
    else
    {
        displayUsage();
        return 0;
    }
}